package dbifc

import DAISY._
import org.scalameter.api._

sealed trait Benchable
case class BenchCase(name: String, annotated: Integer => Unit, unannotated: Integer => Unit,
                     count: Integer = 500) extends Benchable

object DaftBenchmark extends Bench.LocalTime {

  /*
  lazy val executor = LocalExecutor(
    new Executor.Warmer.Default,
    Aggregator.min[Double],
    measurer)
  lazy val measurer = new Measurer.Default
  lazy val reporter = new LoggingReporter[Double]
  lazy val persistor = Persistor.None
   */

  val n = 10
  val ranges: Gen[Int] = Gen.range("size")(n, n * 2, n+1)


  def benchCode(name: String)(code: => Unit): Unit = {
    // println(s"Benchmarking $name")
    performance of name in {
      measure method "bla" in {
        run {
          using(ranges) in { n =>
            Util.cleanDB()
            run { code }
          }
        }
      }
    }
  }
  val i = 10
  // benchCode("Grading[m]") { new Grading().submitAndGradeBench(i) }
  // benchCode("Grading[u]") { new GradingUnannotated().submitAndGradeBench(i) }
  // benchCode("BookNetwork[m]") { new BookNetwork().shareReviewBench(i) }
  // benchCode("BookNetwork[u]") { new BookNetworkUnannotated().shareReviewBench(i) }
  // benchCode("Calendar[m]") { new Calendar().testAddAndViewBench(i) }
  // benchCode("Calendar[u]") { new CalendarUnannotated().testAddAndViewBench(i) }


  def time[R](block: =>R): Long = {
    val t0 = System.nanoTime()
    block
    val t1 = System.nanoTime()
    (t1 - t0)
  }

  def runTestCode(name: String, code: Integer => Unit, count: Integer) = {
    var t: Double = 0
    Util.cleanDB()
    Util.withDB { DB.reset() }
    print(s"Running test $name [$count times]: ")
    run { t = time { code(count) } }
    println(s"${t.toDouble / 1000000000} s")
    t
  }

  def runTest(testcase: Benchable): Unit = testcase match {
    case BenchCase(name, annCode, unannCode, count) =>
      val monTime = runTestCode(s"$name - monitored", annCode, count)
      val unmonTime = runTestCode(s"$name - unmonitored", unannCode, count)
      println(s"Slowdown factor: ${monTime/unmonTime * 100 - 100}%")
  }

  def runBenchmark(): Unit = {
    val base = 5000
    val tests: Seq[Benchable] =
      Seq(BenchCase("Grading",
                    i => new Grading().submitAndGradeBench(i),
                    i => new GradingUnannotated().submitAndGradeBench(i),
                    count=base),
          BenchCase("BookNetwork",
                    i => new BookNetwork().shareReviewBench(i),
                    i => new BookNetworkUnannotated().shareReviewBench(i),
                    count=base/5))
    tests.foreach(runTest)
  }

  runBenchmark()
}
