package dbifc

import DAISY._

@daisy
class BookNetworkNew {
  asAdmin {
    query("CREATE TABLE users (usr)")
    query("CREATE TABLE reviews (id, usr, book, txt, score)")
    query("CREATE TABLE likes (id, usr, revusr, revscore)")
    query("CREATE TABLE friends (userA, userB)")
    query("CREATE TABLE genre(bookid, genre)")
    query("CREATE TABLE stats (genre, revusr, revscore)")
  }

  def addUser(user: String) = {
    query(s"ADD USER $user")
    query(s"INSERT IN users ['$user']")
    query(s"""CREATE VIEW reviews$user(id, book, txt, score) AS OWNER OF
               { b, r, t, s | EXISTS u. ((reviews(b, u, r, t, s) AND u = '$user')) }""")
    query(s"GRANT SELECT ON reviews$user TO $user")
    query(s"GRANT SELECT ON friends TO $user")
    query(s"GRANT INSERT ON reviews TO $user")
    query(s"GRANT SELECT ON stats TO $user")
    val currentUsers = select("{ u | users(u) }")
      .map(x => x.head)
    currentUsers.foreach { user2 =>
       query(s"""CREATE TRIGGER granttig${user}${user2} AFTER INSERT ON friends
     AS OWNER WHEN (userA = '$user' AND userB = '$user2') DO GRANT SELECT ON reviews${user} TO $user2""")
    }
  }

  def addReview(id: String, book: String, review: String, score: String) = {
    val user = whoami()
    query(s"INSERT IN reviews ['$id', '$user', '$book', '$review', '$score']")
  }

  def showMyReviews() = {
    val user = whoami()
    val reviews = query(s"{ b, r, t, s | reviews$user(b, r, t, s) }")
    output(reviews)
  }

  // This should be run by the admin, simulating some kind of
  // friend-request-approval process
  def addFriend(userA: String, userB: String) = {
    query(s"INSERT IN friends ['$userA', '$userB']")
    query(s"INSERT IN friends ['$userB', '$userA']")
  }


  def basicReviewAndRead() = {
    asAdmin {
      addUser("carl")
      addUser("alice")
      addUser("bob")
      addFriend("carl", "alice")
      query(s"GRANT SELECT ON reviewscarl TO alice")
      query(s"GRANT INSERT ON likes TO alice")
    }
    asUser("carl") {
      addReview("a", "WarAndPeace", "Text", "Ten")
      //showMyReviews()
      val reviews = query(s"{ b, r, t, s | EXISTS u. ((reviews(b, u, r, t, s) AND u = 'carl')) }")
      outputTo("carl", reviews)
    }
    asUser("alice") {
      val revs = query(s"{ b, rid, txt, sco | EXISTS u. ((reviews(b, u, rid, txt, sco) AND u = 'carl')) }")
      outputTo("alice", revs)
      query("INSERT INTO likes ['a', 'alice', 'bob', 'Ten']")
    }

    asUser("bob") {
      val friends = select("{ u | EXISTS b. ((friends(b, u) AND b = 'bob')) }")
        .map(_.head)
      val genres = select("{ g, ls | EXISTS b. ((stats(g, b, ls) AND b = 'bob')) }")
        .map(_.head)
      friends.foreach { friend =>
        genres.foreach { genre =>
          outputTo("bob", friend)
          outputTo("bob", genre)
        }
      }
     }
  }

  def reviewWithTrigger() = {
    asAdmin {
      query("INSERT IN stats ['fiction', 'bob', 'one']")
      query(s"""CREATE TRIGGER exampleTrigger AFTER INSERT ON likes
     AS INVOKER
     WHEN EXISTS iid, uusr, book, txt, score. (reviews(id, uusr, book, txt, score))
     DO INSERT IN stats ['fiction', revusr, revscore]""")
    }
    basicReviewAndRead()
  }
  /*
  def shareReviews() = {
    asAdmin {
      addUser("A")
      addUser("B")
      addUser("C")
      addFriend("A", "B")
      addFriend("B", "C")
      addFriend("A", "C")
    }
    asUser("B") {
      addReview("CritiqueOfPureReason", "Great")
    }
    asUser("A") {
      addReview("CritiqueOfPureReason", "PrettyDry")
      val user = whoami()
      val friends = select(s"{ x | EXISTS y. ((friends(y, x) AND y = '$user')) }")
        .map(_.head)
      val reviews = select(s"{ b | EXISTS u,r. ((reviews(b, u, r) AND u = '$user')) }")
        .map(_.head)
      friends.foreach { friend =>
        val friendReviews = select(s"{ b, r | EXISTS u. ((reviews(b, u, r) AND u = '$friend')) }")
        val friendFriends = select(s"{ uu | EXISTS u. ((friends(u, uu) AND u = '$friend' )) }")
          .map(_.head)
        Util.debug(s"Checking mutual friends with $friend: $friendFriends [own friends: $friends]")
        Util.debug(s"Own reviews: $reviews , their reviews: $friendReviews")
        (friendFriends intersect friends).foreach { u2 =>
          friendReviews.foreach { r =>
            Util.debug(s"Checking if $r is a common review")
            if (reviews contains r.head) {
              Util.debug(s"Found common review: $r")
              outputTo(u2, r)
            }
          }
        }
      }
    }
  }

  def shareReviewBench(count: Integer) = {
    asAdmin {
      addUser("A")
      addUser("B")
      addUser("C")
      addFriend("A", "B")
      addFriend("B", "C")
      addFriend("A", "C")
    }
    asUser("B") {
      (1 to count).foreach { n =>
        val alphaNum = Util.numToAlpha(n)
        val book = s"book$alphaNum"
        val score = s"score$alphaNum"
        addReview(book, score)
      }
    }
    asUser("A") {
      (1 to count).foreach { n =>
        val alphaNum = Util.numToAlpha(n)
        val book = s"book$alphaNum"
        val score = s"score$alphaNum"
        addReview(book, score)
        val user = whoami()
        val friends = select(s"{ x | EXISTS y. ((friends(y, x) AND y = '$user')) }")
          .map(_.head)
        val reviews = select(s"{ b | EXISTS u,r. ((reviews(b, u, r) AND u = '$user')) }")
          .map(_.head)
        friends.foreach { friend =>
          val friendReviews = select(s"{ b, r | EXISTS u. ((reviews(b, u, r) AND u = '$friend')) }")
          val friendFriends = select(s"{ uu | EXISTS u. ((friends(u, uu) AND u = '$friend' )) }")
            .map(_.head)
          (friendFriends intersect friends).foreach { u2 =>
            friendReviews.foreach { r =>
              if (reviews contains r.head) {
                outputTo(u2, r)
              }
            }
          }
        }
      }
    }
  }

  def leakReviews() = {
    asAdmin {
      addUser("victim")
      addUser("attacker")
      addFriend("victim", "attacker")
    }
    asUser("victim") {
      addReview("AtlasShrugged", "Trash")
    }
    asUser("attacker") {
      val reviews = select(s"{ b, r | EXISTS u. ((reviews(b, u, r) AND u = 'victim')) }")
      reviews.foreach { review =>
        outputTo("public", review)
      }
    }
  } */

}
