package dbifc

import DAISY._


class BookNetworkUnannotated {
  asAdmin {
    query("CREATE TABLE users (usr)")
    query("CREATE TABLE reviews (book, usr, review)")
    query("CREATE TABLE friends (userA, userB)")
  }

  def addUser(user: String) = {
    query(s"ADD USER $user")
    query(s"INSERT IN users ['$user']")
    query(s"""CREATE VIEW reviews$user(book, review) AS OWNER OF
               { b, r | EXISTS u. ((reviews(b, u, r) AND u = '$user')) }""")
    query(s"GRANT SELECT ON reviews$user TO $user")
    query(s"GRANT SELECT ON friends TO $user")
    query(s"GRANT INSERT ON reviews TO $user")
    val currentUsers = select("{ u | users(u) }")
      .map(x => x.head)
    currentUsers.foreach { user2 =>
       query(s"""CREATE TRIGGER granttig${user}${user2} AFTER INSERT ON friends
     AS OWNER WHEN (userA = '$user' AND userB = 'user2') DO GRANT SELECT ON reviews${user} TO $user2""")
    }
  }

  def addReview(book: String, review: String) = {
    val user = whoami()
    query(s"INSERT IN reviews ['$book', '$user', '$review']")
  }

  def showMyReviews() = {
    val user = whoami()
    val reviews = query(s"{ b, r | reviews$user(b, r) }")
    output(reviews)
  }

  // This should be run by the admin, simulating some kind of
  // friend-request-approval process
  def addFriend(userA: String, userB: String) = {
    query(s"INSERT IN friends ['$userA', '$userB']")
    query(s"INSERT IN friends ['$userB', '$userA']")
  }

  def basicReviewAndRead() = {
    asAdmin {
      addUser("userA")
    }
    asUser("userA") {
      addReview("CritiqueOfPureReason", "PrettyDry")
      showMyReviews()
    }
  }

  def shareReviews() = {
    asAdmin {
      addUser("A")
      addUser("B")
      addUser("C")
      addFriend("A", "B")
      addFriend("B", "C")
      addFriend("A", "C")
    }
    asUser("B") {
      addReview("CritiqueOfPureReason", "Great")
    }
    asUser("A") {
      addReview("CritiqueOfPureReason", "PrettyDry")
      val user = whoami()
      val friends = select(s"{ x | EXISTS y. ((friends(y, x) AND y = '$user')) }")
        .map(_.head)
      val reviews = select(s"{ b | EXISTS u,r. ((reviews(b, u, r) AND u = '$user')) }")
        .map(_.head)
      friends.foreach { friend =>
        val friendReviews = select(s"{ b, r | EXISTS u. ((reviews(b, u, r) AND u = '$friend')) }")
        val friendFriends = select(s"{ uu | EXISTS u. ((friends(u, uu) AND u = '$friend' )) }")
          .map(_.head)
        Util.debug(s"Checking mutual friends with $friend: $friendFriends [own friends: $friends]")
        Util.debug(s"Own reviews: $reviews , their reviews: $friendReviews")
        (friendFriends intersect friends).foreach { u2 =>
          friendReviews.foreach { r =>
            Util.debug(s"Checking if $r is a common review")
            if (reviews contains r.head) {
              Util.debug(s"Found common review: $r")
              outputTo(u2, r)
            }
          }
        }
      }
    }
  }

  def shareReviewBench(count: Integer) = {
    asAdmin {
      addUser("A")
      addUser("B")
      addUser("C")
      addFriend("A", "B")
      addFriend("B", "C")
      addFriend("A", "C")
    }
    asUser("B") {
      (1 to count).foreach { n =>
        val alphaNum = Util.numToAlpha(n)
        val book = s"book$alphaNum"
        val score = s"score$alphaNum"
        addReview(book, score)
      }
    }
    asUser("A") {
      (1 to count).foreach { n =>
        val alphaNum = Util.numToAlpha(n)
        val book = s"book$alphaNum"
        val score = s"score$alphaNum"
        addReview(book, score)
        val user = whoami()
        val friends = select(s"{ x | EXISTS y. ((friends(y, x) AND y = '$user')) }")
          .map(_.head)
        val reviews = select(s"{ b | EXISTS u,r. ((reviews(b, u, r) AND u = '$user')) }")
          .map(_.head)
        friends.foreach { friend =>
          val friendReviews = select(s"{ b, r | EXISTS u. ((reviews(b, u, r) AND u = '$friend')) }")
          val friendFriends = select(s"{ uu | EXISTS u. ((friends(u, uu) AND u = '$friend' )) }")
            .map(_.head)
          (friendFriends intersect friends).foreach { u2 =>
            friendReviews.foreach { r =>
              if (reviews contains r.head) {
                outputTo(u2, r)
              }
            }
          }
        }
      }
    }
  }

  def leakReviews() = {
    asAdmin {
      addUser("victim")
      addUser("attacker")
      addFriend("victim", "attacker")
    }
    asUser("victim") {
      addReview("AtlasShrugged", "Trash")
    }
    asUser("attacker") {
      val reviews = select(s"{ b, r | EXISTS u. ((reviews(b, u, r) AND u = 'victim')) }")
      reviews.foreach { review =>
        outputTo("public", review)
      }
    }
  }

}
