package dbifc

import DAISY._


class CalendarUnannotated {
  asAdmin {
    query("CREATE TABLE events(eventName, date, owner)")
  }

  def addUser(user: String) = {
    query(s"ADD USER $user")
    query(s"GRANT INSERT ON events TO $user")
  }

  def createEvent(name: String, date: String) = {
    val user = whoami()
    query(s"INSERT IN events ['$name', '$date', '$user']")
    query(s"CREATE TABLE event${name}(attendee)")
    query(s"INSERT IN event${name} ['$user']")
  }

  def addAttendee(name: String, attendee: String) = {
    query(s"INSERT IN event${name} ['$attendee']")
    query(s"GRANT SELECT ON event${name} TO $attendee")
  }

  def removeAttendee(name: String, attendee: String) = {
    query(s"DELETE FROM event${name} ['Bob']")
    query(s"REVOKE SELECT ON event${name} FROM $attendee")
  }

  def getAttendees(name: String): Set[String] = {
    select(s"{a | event${name}(a)}").map(_.head)
  }

  def viewAttendees(name: String) = {
    val attendees = getAttendees(name)
    output(attendees)
  }

  def testAddAndView() = {
    asAdmin {
      addUser("Alice")
      addUser("Bob")
      addUser("Charlie")
    }
    asUser("Alice") {
      createEvent("KeySigningParty", "xyz")
      addAttendee("KeySigningParty", "Bob")
      addAttendee("KeySigningParty", "Charlie")
    }
    asUser("Charlie") {
      viewAttendees("KeySigningParty")
    }
  }

  def testLeakParticipants() = {
    asAdmin {
      addUser("Alice")
      addUser("Bob")
      addUser("Charlie")
    }
    asUser("Alice") {
      createEvent("KeySigningParty", "xyz")
      addAttendee("KeySigningParty", "Bob")
      addAttendee("KeySigningParty", "Charlie")
    }
    asUser("Charlie") {
      val attendees = getAttendees("KeySigningParty")
      outputTo("public", attendees)
    }
  }

  def viewAttendeesAfterRemoved() = {
    asAdmin {
      addUser("Alice")
      addUser("Bob")
      addUser("Charlie")
    }
    asUser("Alice") {
      createEvent("KeySigningParty", "xyz")
      addAttendee("KeySigningParty", "Bob")
      addAttendee("KeySigningParty", "Charlie")
    }
    asUser("Charlie") {
      viewAttendees("KeySigningParty")
    }
    asUser("Alice") {
      removeAttendee("KeySigningParty", "Bob")
    }
    asUser("Bob") {
      viewAttendees("KeySigningParty")
    }
  }
}
