package dbifc

import prototype.database.postgresql._
import prototype.system.postgresql._
import prototype.system.interfaces._
import prototype.system.implementation._
import prototype.users.{UserFactory}
import prototype.statements.implementation.postgresql._
import prototype.statements.abstractImplementation._
import prototype.parser.StatementParser
import prototype.util._
import prototype.pdp.AlwaysAuthorizedPdp

import Types._
import java.io.ByteArrayOutputStream

class DB {

  private val dbAdmin = "dbifc"
  private val dbPasswd = "abcdef"
  private val dbURL = "localhost:5432/dbifc"
  val conf = new PostgreSqlConfiguration(dbURL, dbAdmin, dbPasswd)
  val man = new PostgreSqlConnectionManager(conf)
  val stateFactory: PostgreSqlStateFactory =
    new PostgreSqlStateFactory()
  val engine = new PostgreSqlDatabaseEngine(conf, stateFactory)
  val statementFactory: IStatementFactory =
    PostgreSqlStatementFactory.getInstance
  val userFactory = UserFactory.getInstance
  val parser = StatementParser.getInstance

  val logger = new Logger(new java.io.File("logs.txt"), true)

  val dac = new AlwaysAuthorizedPdp()
  val system: ISystem = new System(dac, engine, statementFactory, logger)
  def state = system.getSystemState
  // println(s"Database initialized (admin: ${userFactory.getAdmin})")

  // TODO: avoid sql-injection attacks using the user name
  def parseQuery(u: User, queryString: String): SQLStmt = {
    val boas = new ByteArrayOutputStream()
    val oldOut: java.io.PrintStream = java.lang.System.out
    val oldErr: java.io.PrintStream = java.lang.System.err
    java.lang.System.setOut(new java.io.PrintStream(boas))
    java.lang.System.setErr(new java.io.PrintStream(boas))
    val res = parser.create(s"AS $u: $queryString", statementFactory, userFactory, system.getSystemState)
    java.lang.System.setOut(oldOut)
    java.lang.System.setErr(oldErr)
    res
  }
}

object DB {
  val x = 5
  var v = new DB()
  def reset() = {
    NaiveDAC.resetPolicy()
    try {
      v.engine.close()
      v.man.close()
    } catch {
      case e: Exception =>
        // throw new IllegalAccessError("Failed to close database connection")
    }
    v = new DB()
  }

}
