package dbifc

import DAISY._

class GradingUnannotated {

  val r = scala.util.Random

  asAdmin {
    query("CREATE TABLE students(sname)")
    query("CREATE TABLE studentcodes(sname,scode)")
    query("CREATE TABLE tas(tname)")
    query("CREATE TABLE codeHandins(scode, handinId)")
    query("CREATE TABLE handins(handinId, handinText)")
    query("CREATE TABLE feedback(handinId, grade)")
  }

  def addStudent(student: String) = {
    val code = math.abs(r.nextInt)
    query(s"ADD USER $student")
    query(s"INSERT IN students ['$student']")
    query(s"INSERT IN studentcodes ['$student', 's$code']")
    query(s"""CREATE VIEW code${student}(code) AS OWNER OF
               { x | EXISTS y. ((studentcodes(y, x) AND y = '$student')) }""")
    query(s"""GRANT SELECT ON code${student} TO $student""")

    query(s"GRANT INSERT ON handins TO $student")
    query(s"GRANT INSERT ON codeHandins TO $student")

    query(s"""CREATE VIEW feedback${student}(handinId, grade) AS OWNER OF
               { x, y | EXISTS z. ((feedback(x, y) AND z = 's$code' AND codeHandins(z, x))) }""")
    query(s"""GRANT SELECT ON feedback${student} TO $student""")
  }

  def addTA(ta: String) = {
    query(s"ADD USER $ta")
    query(s"INSERT IN tas ['$ta']")
    query(s"GRANT SELECT ON handins TO $ta")
    query(s"GRANT INSERT ON feedback TO $ta")
    query(s"GRANT SELECT ON feedback TO $ta")
  }

  def submitHandin(handinText: String) = {
    val student: String = whoami()
    val code = query(s"{ c | code${student}(c) }")
      .asInstanceOf[Set[Seq[String]]].head.head
    // FIXME: Make sure handinId is unique
    val id: String = math.abs(r.nextInt).toString
    val handinId = s"submission${id}"
    query(s"INSERT IN handins [ '$handinId', '$handinText' ]")
    query(s"INSERT IN codeHandins [ '$code', '$handinId' ]")
  }

  def viewSubmissions(): Set[Seq[String]] = {
    val result = select("{ handinId, handinText | handins(handinId, handinText) }")
    result
  }

  def grade(handinId: String, feedback: String) = {
    query(s"INSERT IN feedback ['$handinId', '$feedback']")
  }

  def viewFeedback() = {
    val student: String = whoami()
    output(query(s"{ x, y | feedback$student(x, y) }"))
  }

  asAdmin {
    addStudent("studentA")
    addTA("taA")
  }

  def setup() = {
    asUser("studentA") {
      submitHandin("prettyGoodSubmission")
    }
    asUser("taA") {
      val firstSubmission = viewSubmissions().head
      // TA inspects submissionA, decides to give it feedback 'prettyGood'
      grade(firstSubmission.head, "prettyGood")
    }
  }

  def submitAndGrade() = {
    setup()
    asUser("studentA") {
      // student looks at feedback
      viewFeedback()
    }
  }

  def submitAndGradeBench(count: Integer) = {
    (1 to count).foreach { n =>
      val nStr = Util.numToAlpha(n)
      asUser("studentA") {
        submitHandin(s"submission$nStr")
      }
      asUser("taA") {
        val submission = viewSubmissions().head
        grade(submission.head, s"pretty$nStr")
      }
    }
  }

  def accessStudentCodesAsTA() = {
    asUser("taA") {
      val x = query("{ (s, c) | studentcodes(s, c) }")
      output(x)
    }
  }

  def leakFeedback() = {
    setup()
    asUser("taB") {
      // and tries to leak everything
      val feedbacks = query("{ id, grade | feedback(id, grade) }")
      outputTo("public", feedbacks)
    }
  }

  def newTAreadsFeedback() = {
    setup()
    // new TA joins the team
    asAdmin {
      addTA("taB")
    }
    asUser("taB") {
      // and outputs information to themselves
      val feedbacks = query("{ id, grade | feedback(id, grade) }")
      outputTo("taB", feedbacks)
    }
  }

  def studentReadingOtherSubmissions() = {
    addStudent("studentB")
    asUser("studentB") {
      val submissions = query("{ i, r | handins(i, r) }")
      output(submissions)
    }
  }

}
