package dbifc

import scala.io.Source
import sys.process._
import scala.collection.JavaConverters._
import java.io.{File, PrintWriter}

import Types._

import prototype.database.postgresql._
import prototype.system.postgresql._
import prototype.pdp._

case class Config(target: String = "test.w",
                  interpret: Boolean = false,
                  euroSPDAC: Boolean = false,
                  internal: Boolean = false,
                  adjust: Boolean = false)

object Main extends Object {

  val toolName = "dbifc"
  val homeDir = System.getProperty("user.home")
  val parser = new scopt.OptionParser[Config](toolName) {
    head(toolName, "0.1")
    arg[String]("<target>") optional() action { (x, c) => c.copy(target = x) } text(
      "Target program")
    opt[Unit]("interpret") abbr("i") action { (_, c) => c.copy(interpret = true) } text(
      "Interpret the program without security monitoring.")
    opt[Unit]("internal") abbr("I") action { (_, c) => c.copy(internal = true) } text(
      "Use internal attacker model")
    help("help") abbr("h") text("Print this message")
  }

  def clean() = {
    /*
    val exitCode = Seq("psql", "-d", "dbifc", "-U", "admin",
                       "-c", "DROP TABLE __adom__expr__values__ CASCADE ;").!
     */
  }
  /*
  clean()
  parser.parse(args, Config()) match {
    case None =>
    case Some(cfg) => run(cfg)
  }*/

  var options: Config = Config()

  def newDAC(): DACMechanism = {
    if (options.euroSPDAC) {
      DACMechanism.pdpToDAC(
        new ConfidentialityAndIntegrityPdp(DB.v.stateFactory, DB.v.engine,
                                           DB.v.userFactory,
                                           DB.v.statementFactory, DB.v.logger))
    } else {
      NaiveDAC
    }
  }

}
