package dbifc

import DAISY.{query, changeUser, asAdmin, output}

/* Case study 1 */

@daisy
class PerUserSecrets {
  def newUser(user: String) = {
    asAdmin {
      // attacker exists by default in our db setup:
      if (user != "attacker") {
        query(s"ADD USER $user")
      }
      query(s"INSERT IN secrets ['$user', '${user}Secret']")
      query(s"""CREATE VIEW secrets${user} (secret) AS OWNER
               OF { x | EXISTS y. ((secrets(y, x) AND y = '$user')) }""")
      query(s"CREATE TRIGGER ${user}Authorized AFTER INSERT ON authorized AS INVOKER WHEN usr = '$user' DO GRANT SELECT ON secrets${user} TO ${user}")
    }
  }

  def authorize(user: String) = {
    query(s"INSERT IN authorized ['$user']")
  }

  asAdmin {
    query("CREATE TABLE secrets(usr, secret)")
    query("CREATE TABLE authorized(usr)")
  }
  newUser("attacker")
  newUser("honest")


  def outputSecrets(user: String) = {
    var res: Any = null
    res = query(s"{x | secrets${user}(x)}")
    output(res)
  }

  def testAttackerAuthorized() = {
    asAdmin {
      authorize("attacker")
    }
    outputSecrets("attacker")
  }

  def testAttackerUnauthorized() = {
    outputSecrets("attacker")
  }

  def testAttackerReadOthersAuthorized() = {
    asAdmin {
      authorize("attacker")
    }
    outputSecrets("honest")
  }

  def testSelfAuthorize() = {
    changeUser("attacker")
    authorize("attacker")
  }
}
