package dbifc

import java.io.{File, FileOutputStream, PrintWriter}
import scala.collection.mutable.StringBuilder
import sys.process._
import scala.util.matching._
import scala.language.implicitConversions
import scala.io.Source

//import scala.collection._
//import scala.collection.immutable._

import prototype.util.Pair
import prototype.users.IUser

import Types._

object Util {
  var sound: Boolean = true
  var indent: Integer = 0
  var nsu: Boolean = true

  var muffle: Boolean = true

  def debug(msg: Object): Unit = {
    if (!muffle) {
      println(s"${"  " * Util.indent}$msg")
    }
  }

  def muffled[T](code: =>T): T = {
    val oldMuffle = muffle
    muffle = true
    val res = code
    muffle = oldMuffle
    res
  }

  def warn(msg: Object): Unit = debug(Console.YELLOW + s"WARN: $msg" + Console.RESET)

  def error(msg: Object, pref: String = "ERR: "): Unit = debug(Console.RED + s"$pref$msg" + Console.RESET)

  def die(o: Object, exitCode: Integer = 2): Nothing = {
    error(o, "")
    throw new IllegalArgumentException(o.toString)
  }

  def setIndent(i: Integer): Unit = { Util.indent = i }

  def incIndent(): Unit = setIndent(indent + 1)

  def decIndent(): Unit = setIndent(scala.math.max(indent - 1, 0))

  def indented(act: =>Unit): Unit = {
    incIndent()
    act
    decIndent()
  }

  def unsound(): Unit = { sound = false }

  def repeat[A](body: => A) = new {
    def until(condition: A => Boolean): A = {
      var a = body
      while (!condition(a)) { a = body }
      a
    }
  }

  implicit def pairToScala[A, B](p: Pair[A, B]): (A, B) =
    (p.getFirst, p.getSecond)

  implicit def scalaToPair[A, B](p: (A, B)): Pair[A, B] =
    new Pair(p._1, p._2)

  implicit def iuserToUser(u: IUser): User = u.getUsername

  def withDB[A](act: => A): A = {
    try {
      act
    } finally {
      DB.v.engine.close()
    }
  }

  // Since the underlying library doesn't support numbers
  // in tuple values, we encode numbers alphabetically:
  val digits = (0 to 10).map(n => (0x41 + n).toChar).toArray
  def numToAlpha(n: Integer): String = {
    var i = n
    var str = new StringBuilder
    while (i > 0) {
      val digit = i % 10;
      i = i / 10;
      str += digits(digit);
    }
    str.toString
  }

  def cleanDB() = {
    val tables = Seq("__adom__expr__values__", "codehandins", "feedback", "handins",
                     "studentcodes", "students", "tas", "friends", "users", "reviews")
    tables.foreach { table =>
      val ec = Seq("psql", "-d", "dbifc", "-U", "dbifc",
                   "-c", s"DROP TABLE $table CASCADE ;").!(ProcessLogger(_ => ()))
      debug(s"Deleting table $table returned exit code $ec")
    }
  }

  def debugFile(o: Object) = {
    scala.tools.nsc.io.File("log.txt").appendAll(o.toString + "\n")
  }
}
