package dbifc
import scala.reflect.macros.Context
import scala.language.experimental.macros
import scala.annotation.StaticAnnotation

import scala.collection.immutable.Set

class daisyMacroHelper[U <: scala.reflect.api.Universe]
  (val c: Any, val u: U) {
  import u._

  val monitorObject = q"dbifc.Monitor.mon"
  //val bottomLevel = q"dbifc.Bottom()"
  //val topLevel = q"dbifc.Top()"
  //val bottomLevel = q"$monitorObject.getCfg.top"
  //val topLevel = q"$monitorObject.getCfg.bot"
  val bottomLevel: Tree = q"$monitorObject.bot"
  val topLevel: Tree = q"$monitorObject.top"
  val userLevel = q"$monitorObject.userLevel"

  def debug(s: Any): Unit = {
    c match {
      case cxt: Context =>
        cxt.info(cxt.enclosingPosition, s"$s", true)
      case _ =>
        println(s"Debug: $s")
    }
  }

  def error(s: Any): Nothing = {
    c match {
      case cxt: Context =>
        cxt.error(cxt.enclosingPosition, s"$s")
        throw new RuntimeException("Transformation failed")
      case _ =>
        throw new RuntimeException(s"$s")
    }
  }

  def freeVariables(t: Tree): Set[TermName] = {
    class FreeVariableTraverser extends Traverser {

      var vars: Set[TermName] = Set()
      override def traverse(tree: Tree): Unit = tree match {
        case Ident(x: TermName) =>
          vars = vars ++ Set(x.asInstanceOf[TermName])
          super.traverse(tree)
        case _ => super.traverse(tree)
      }
    }
    val trav = new FreeVariableTraverser()
    trav.traverse(t)
    trav.vars
  }

  def freeVariablesList(expressions: Seq[Tree]): Set[TermName] = {
    expressions.flatMap(expr => freeVariables(expr)).toSet
  }

  def joinLevelsTerm(xs: Set[String]): Tree = {
    xs.foldRight(q"$bottomLevel")((x, t) =>
      q"$monitorObject.getCfg.join($monitorObject.lbls.getLevel($x), $t)")
  }

  def termToString(x: TermName): String = x.encodedName.toString

  def exprLevel(e: Tree): Tree =
    joinLevelsTerm(freeVariables(e).map(termToString))

  def exprsLevel(es: Seq[Tree]): Tree = {
    es.foldRight(q"$bottomLevel")((e, t) =>
      q"$monitorObject.getCfg.join(${exprLevel(e)}, $t)")
  }

  def makePopLevelsCall(funSig: (String, Int),
                        paramsList: List[String]): Tree = {
    q"""
($paramsList, $monitorObject.popCallLevels((${funSig._1}, ${funSig._2}))).zipped.map((param, level) =>
 $monitorObject.setLevel(param, level))"""
  }

  def setLevelWithPC(x: TermName, l: Tree): Tree = {
    q"""$monitorObject.setLevel(${termToString(x)},
                                $monitorObject.getCfg.join($monitorObject.getPC,
                                                           $l))"""
  }

  def freshName(prefix: String): TermName = {
    u.internal.reificationSupport.freshTermName(prefix)
  }

  def assignmentContext(x: TermName, rhs: Tree, stmt: Tree): Tree = {
    val xName = termToString(x)
    // FIXME: this doesn't allow using assignments as values at the moment
    val xLevel = freshName(xName ++ "level")
      q"""
val $xLevel = $monitorObject.getLevel($xName)
if(!($monitorObject.getCfg.le($monitorObject.getPC, $xLevel))) {
  $monitorObject.securityException("NSU check failed: assigning to variable of level " ++ $xLevel.toString ++ " in context " ++ $monitorObject.getPC.toString)
}
$monitorObject.newLastQueryLabel()
$monitorObject.newFunctionReturnScope()
$stmt
${setLevelWithPC(x, exprLevel(rhs))}
$monitorObject.upgradeLevel($xName, $monitorObject.popLastQueryLabel())
$monitorObject.upgradeLevel($xName, $monitorObject.popFunctionReturnScope())
"""
  }

  def transformExpr(path: Seq[String], expr: Tree): Tree = {
    expr match {
      case q"" => q""
      case Literal(lit) =>
        q"$lit"
      case q"$e1 = $e2" =>
        val e2transformed = transformExpr(path, e2)
        e1 match {
          case q"${x: TermName}" =>
            assignmentContext(x, e2, q"$e1 = $e2transformed")
          case _ =>
            error(s"Unhandled term on LHS of assignment: $e1")
        }
      case q"$mods val $pat: $tpt = $rhs" =>
        val transformedRHS = transformExpr(path, rhs)
        debug(s"Found val def $pat = $rhs (${showRaw(expr)})")
        // q"$mods val $pat = $rhs"
        // ValDef(mods, pat, rhs)
        val resultVar = freshName("result")
        pat match {
          case tname: TermName =>
            val xName = termToString(tname)
            q"""$mods val $pat: $tpt = {
$monitorObject.newLastQueryLabel()
$monitorObject.newFunctionReturnScope()
val $resultVar = $transformedRHS
${setLevelWithPC(xName, exprLevel(rhs))}
$monitorObject.upgradeLevel($xName, $monitorObject.popLastQueryLabel())
$monitorObject.upgradeLevel($xName, $monitorObject.popFunctionReturnScope())
 $resultVar }"""
          case _ =>
            debug(s"Unhandled value definition $expr")
            expr
        }
      case q"if ($cond) $thn else $els" =>
        val backupTerm = freshName("oldPC")
        q"""val $backupTerm = $monitorObject.getPC ;
              $monitorObject.upgradePC(${exprLevel(cond)}) ;
              if ($cond) ${transformExpr(path, thn)}
                 else ${transformExpr(path, els)} ;
              $monitorObject.restorePC($backupTerm)"""
      case q"while ($cond) $body" =>
        val backupPC = freshName("oldPC")
        q"""val $backupPC = $monitorObject.getPC ;
              $monitorObject.upgradePC(${exprLevel(cond)}) ;
              while ($cond) { ${transformExpr(path, body)} } ;
              $monitorObject.restorePC($backupPC)"""
      case q"for (..$enums) $expr" =>
        val backupPC = freshName("oldPC")
        q"""val $backupPC = $monitorObject.getPC
            $monitorObject.upgradePC(${exprsLevel(enums)})
            for (..${enums.map(transformExpr(path, _))}) ${transformExpr(path, expr)}
            $monitorObject.restorePC($backupPC)
         """
      case q"$matchedExpr match { case ..$cases }" =>
        val backupPC = freshName("oldPC")
        val matchedTransformed = transformExpr(path, matchedExpr)
        val transformedCases = cases.map(transformExpr(path, _))
        q"""val $backupPC = $monitorObject.getPC
            $monitorObject.upgradePC(${exprLevel(matchedExpr)})
            $matchedTransformed match {
               case ..$transformedCases
            }
            $monitorObject.restorePC($backupPC)
        """
      case q"$objExpr.foreach { $argname => $expr }" =>
        debug(s"Found call to foreach!")
        val backupPC = freshName("oldPC")
        val result = freshName("result")
        q"""
$objExpr.foreach { $argname =>
  $monitorObject.upgradeLevel(${termToString(argname.name)}, ${exprLevel(objExpr)})
  ${transformExpr(path, expr)}
}
"""
      case (obj @ Apply(f, argLists)) =>
        // TODO: handle chained application
        val transformedArgLists = argLists.map(argList =>
          transformExpr(path, argList))
        val arity = argLists.size
        val funSig = (f.toString, arity)
        val argumentLevels = argLists.map(arg => exprLevel(arg))
        val pushArgumentLevelExpr: Tree =
          q"$monitorObject.pushCallLevels($funSig, Seq(..$argumentLevels))"
        debug(s"Function call $f,  ${f.symbol}, owner: ${f.symbol.owner}")
        if (f == q"DAISY.query" || f.toString == "query") {
          val fvs = freeVariablesList(argLists)
          debug(s"Found query: $argLists with free vars $fvs")
          val fvsStrings = fvs.map(tn => tn.toString)
          Apply(q"DAISY.expandQuery", transformedArgLists ++ List(q"$fvsStrings"))
        } else {
          val callExpr = Apply(f, transformedArgLists)
          q"""$pushArgumentLevelExpr
$callExpr"""
        }
      case Block(exprs, tree) =>
        q"""..${exprs.map(expr1 => transformExpr(path, expr1))} ;
        ${transformExpr(path, tree)}"""
      case _ =>
        debug(s"Unhandled expression $expr [${showRaw(expr)}]")
        expr
    }
  }

  def transformDef(path: Seq[String], x : Tree): Tree = {
    x match {
      case (obj @ q"$mods def $tname[..$tparams](...$paramss): $tpt = $expr") =>
        debug(s"Found definition $tname with args $paramss and body $expr")
        // TODO: check scoping related to defs:
        if (paramss.length > 1) {
          error(s"curried applications not supported for the moment: $obj")
        }
        val transformedBody = transformExpr(path, expr)
        if (paramss.length == 1) {
          val arity = paramss.head.length
          val funSig = (termToString(tname), arity)
          val paramsList = paramss.head.map(_.name.toString)
          val popLevelsCall = makePopLevelsCall(funSig, paramsList)
          val resVar = freshName("result")
          q"""$mods def $tname[..$tparams](...$paramss): $tpt = {
$monitorObject.newLevelScope()
$popLevelsCall
val $resVar = $transformedBody
$monitorObject.upgradeFunctionReturnLevel(${exprLevel(transformedBody)})
$monitorObject.leaveLevelScope()
$resVar
}"""
        } else {
          q"$mods def $tname[..$tparams](...$paramss): $tpt = $transformedBody"
        }
      case e: TermTree =>
        transformExpr(Seq(), e)
      case x =>
        debug(s"Unhandled construct ${showRaw(x)}")
        x
    }
  }

  def transformToplevelObj(obj: Tree): Tree = {
    obj match {
      case q"object $name extends ..$parents { ..$body }" =>
        debug(s"Found object definition $body")
        q"""object $name extends ..$parents {
            ..${body.map(elt => transformDef(Seq(name.toString), elt))}
          }
          """
      case q"$mods class $tpname[..$tparams] $ctorMods(...$paramss) extends { ..$earlydefns } with ..$parents { $self => ..$stats }" =>
        debug(s"Found class definition $stats (self = $self)")
        q"""$mods class $tpname[..$tparams] $ctorMods(...$paramss) extends { ..$earlydefns } with ..$parents { $self => ..${stats.map(elt => transformDef(Seq(tpname.toString), elt))} }"""
      case obj: TermTree => transformExpr(Seq(), obj)
      case obj: DefDef => transformDef(Seq(), obj)
      case x =>
        error(s"Cannot handle annotated form $x")
    }
  }

  def transformSink(sink: Seq[Tree]): Seq[Tree] = {
    sink match {
      case (obj @ q"$mods def $tname[..$tparams](...$paramss): $tpt = $expr") :: Nil =>
        val pcCheck = freshName("pcCheck")
        val valCheck = freshName("valCheck")
        /*
        val cxt: Context = c match {
          case cxt: Context => cxt
          case _ => error("transformSink called at runtime")
        }
        val funName = cxt.enclosingClass match {
          case q"""$mods object $objName extends { ..$earlydefns }
              with ..$parents { $self => ..$stats }""" =>
            s"$objName.$tname"
          case _ => error("Sink not enclosed in object")
            // FIXME: support more locations for sinks
        } */
        val paramNames = paramss.flatMap(_.map(valdef => valdef.name.toString))
        val valCheckTerm = joinLevelsTerm(paramNames.toSet)
        val popLevelsCall = if (paramss.length == 1) {
          makePopLevelsCall((termToString(tname), paramss.head.length), paramNames)
        } else {
          q""
        }
        q"""$mods def $tname[..$tparams](...$paramss): $tpt = {
$popLevelsCall
val $pcCheck = $monitorObject.getCfg.le($monitorObject.getPC, $userLevel)
val $valCheck = $monitorObject.getCfg.le($valCheckTerm, $userLevel)
if($valCheck) {
  if($pcCheck) {
     $expr
  } else {
    $monitorObject.securityException("Low output occured in high context")
  }
} else {
  $monitorObject.securityException("Attempt to output data at level " ++ $valCheckTerm.toString ++
  "on to user at level " ++ $userLevel.toString)
} }""" :: Nil
      case _ =>
        error("Invalid target for sink annotation")
        ???
    }
  }

  def transformSink2(sink: Seq[Tree]): Seq[Tree] = {
    sink match {
      case (obj @ q"$mods def $tname[..$tparams](...$paramss): $tpt = $expr") :: Nil =>
        val pcCheck = freshName("pcCheck")
        val valCheck = freshName("valCheck")
        val paramNames = paramss.flatMap(_.map(valdef => valdef.name.toString))
        if (paramss.length != 1 || paramss.head.length != 2) {
          error(s"@sink2 needs a 2-argument function")
        }

        val valCheckTerm = joinLevelsTerm(Set(paramss.head.apply(1).name.toString))
        val popLevelsCall = if (paramss.length == 1) {
          makePopLevelsCall((termToString(tname), paramss.head.length), paramNames)
        } else {
          q""
        }
        val targetUserLevel = q"$monitorObject.getCfg.userLevel(${paramss.head.head.name})"
        q"""$mods def $tname[..$tparams](...$paramss): $tpt = {
$popLevelsCall
val $pcCheck = $monitorObject.getCfg.le($monitorObject.getPC, $targetUserLevel)
val $valCheck = $monitorObject.getCfg.le($valCheckTerm, $targetUserLevel)
if($valCheck) {
  if($pcCheck) {
     $expr
  } else {
    $monitorObject.securityException("Low output occured in high context")
  }
} else {
  $monitorObject.securityException("Attempt to output data at level " ++ $valCheckTerm.toString ++
  "on to user at level " ++ $targetUserLevel.toString)
} }""" :: Nil
      case _ =>
        error("Invalid target for sink annotation")
        ???
    }
  }

  def transformCode(code: Tree): Tree = {
    debug(s"Transforming expression $code [${code.getClass}]")
    transformExpr(Seq(), code)
  }

  def transformHigh(code: Seq[Tree]): Seq[Tree] = {
    code match {
      case (obj @ q"$mods val $tname: $tpt = $expr") :: Nil =>
        val res = Seq(q"""$obj""",
                      q"$monitorObject.setLevel(${termToString(tname)}, $topLevel)")
        debug(s"Adding high decl: $res")
        res
      case (obj @ q"$mods var $tname: $tpt = $expr") :: Nil =>
        Seq(q"$obj", q"$monitorObject.setLevel(${termToString(tname)}, $topLevel)")
      case _ =>
        error(s"@high cannot handle construct $code [${code.getClass}]")
    }
  }

}

object daisyMacro {

  def test() = println("Callable from main project!")

  def impl(c: Context)(annottees: c.Expr[Any]*): c.Expr[Any] = {
    import c.universe._

    val helper = new daisyMacroHelper[c.universe.type](c, c.universe)
    val transformed =
      annottees.map(_.tree)
        .toList.map(obj => helper.transformToplevelObj(obj))
    val result = q"..$transformed"
    c.Expr[Any](result)
  }

  def sinkImpl(c: Context)(annottees: c.Expr[Any]*): c.Expr[Any] = {
    import c.universe._
    val helper = new daisyMacroHelper[c.universe.type](c, c.universe)
    val result = helper.transformSink(annottees.map(_.tree).toList)
    c.Expr[Any](q"..$result")
  }

  def sink2Impl(c: Context)(annottees: c.Expr[Any]*): c.Expr[Any] = {
    import c.universe._
    val helper = new daisyMacroHelper[c.universe.type](c, c.universe)
    val result = helper.transformSink2(annottees.map(_.tree).toList)
    c.Expr[Any](q"..$result")
  }

  def daisyImpl(c: Context)(code: c.Expr[Any]): c.Expr[Any] = {
    import c.universe._
    val helper = new daisyMacroHelper[c.universe.type](c, c.universe)
    val result = helper.transformCode(code.tree)
    c.Expr[Any](q"$result")
  }

  def daisy(code: Any): Any = macro daisyImpl

  def highImpl(c: Context)(annottees: c.Expr[Any]*): c.Expr[Any] = {
    import c.universe._
    val helper = new daisyMacroHelper[c.universe.type](c, c.universe)
    val result = helper.transformHigh(annottees.map(_.tree).toList)
    c.Expr[Any](q"..$result")
  }
}

class daisy extends StaticAnnotation {
  def macroTransform(annottees: Any*): Any = macro daisyMacro.impl
}

class sink extends StaticAnnotation {
  def macroTransform(annottees: Any*): Any = macro daisyMacro.sinkImpl
}

class sink2 extends StaticAnnotation {
  def macroTransform(annottees: Any*): Any = macro daisyMacro.sink2Impl
}

class high extends StaticAnnotation {
  def macroTransform(annottees: Any*): Any = macro daisyMacro.highImpl
}
